from datetime import datetime
from operator import attrgetter
from dela.file_reader import FileReader
from dela.todo import Todo
from dela.logger import log


class TodoQueryContext:
    _all = []
    _result = []


class TodoQuery:
    reader = FileReader

    def __init__(self, glob):
        self.glob = glob
        self.context = TodoQueryContext()
        self.context._all = self._read_fs()

    def _read_fs(self):
        result = []
        for line in self.reader.get_lines(self.glob):
            todo = Todo.from_line(line.content, line.file, line.number)
            if todo:
                todo.merge(line.file_meta)
                result.append(todo)

        return result

    def start(self):
        self.context._result = self.context._all
        return self

    def filter(self, filter_fn):
        """
        Generic filter against current result
        """
        self.context._result = [
            t for t in self.context._result if filter_fn(t)
        ]
        return self

    def status_in(self, statuses):
        """
        Filter out all todos with a status not in `statuses`
        """
        self.filter(lambda todo: todo.status in statuses)
        return self

    def tag_in(self, tags):
        """
        Filter out all todos haven't tags intersect with given `tags`
        """
        self.filter(lambda todo: bool([t for t in todo.tags if t in tags]))
        return self

    def exclude_upcoming(self, today=int(datetime.now().strftime('%Y%m%d'))):
        """
        Filter out all upcoming todos
        """
        self.filter(
            lambda todo: not bool(todo.date) or int(todo.date) <= today
        )
        return self

    def today(self, today=int(datetime.now().strftime('%Y%m%d'))):
        """
        Filter out all todos except today ones
        """
        self.filter(lambda todo: bool(todo.date) and int(todo.date) <= today)
        return self

    def sort(self, comparator):
        self.context._result = comparator(self.context._result)
        return self

    def sort_by(self, attr, reverse=False):
        self.context._result = sorted(
            self.context._result,
            key=attrgetter(*attr if list == type(attr) else attr),
            reverse=reverse,
        )
        return self

    def sort_by_statuses(self, statuses_order, reverse=False):
        order = {key: i for i, key in enumerate(statuses_order, start=1)}
        self.context._result = sorted(
            self.context._result,
            key=lambda t: order[getattr(t, 'status')],
            reverse=reverse,
        )
        return self

    def exec(self):
        result = self.context._result
        self.context._result = []
        return result



import codecs
from string import Template


class TodoPresentation:
    def __init__(self, format) -> None:
        self.template = Template(format)

    def present(self, todos):
        for todo in todos:
            self.present_one(todo)

    def present_one(self, todo):
        context = todo.__dict__
        context['tags'] = ' '.join(context['tags'])

        str = self.template.substitute(context)

        # NOTE: This 2 lines are here to add support for unicode escape
        #       characters like terminal colors, tabs, etc
        str = codecs.decode(str, 'raw-unicode-escape')
        str = str.encode('latin1').decode('utf8')

        print(str)



#! /bin/python
import pkg_resources
from docopt import docopt
from dela.list_comand_config import ListCommandConfig
from dela.list_command import ListCommand
from dela.logger import log
from dela.todo_presentation import TodoPresentation

doc = """dela

CLI to list todos from markdown files and obsidian vaults.

Usage:
  dela -h | --help
  dela --version
  dela list [-v] [--today|--upcoming] [--all|--statuses=<str>] [--tags=<str>] [--sort_by=<str>] [--format=<str>] [<glob>]


Options:
  -h --help                     Show this screen
  -v --verbose                  Enable logging
  --version                     Show version.

  -s=<str> --statuses=<str>     Filter by status [default: ~,n, ]
  -t=<str> --tags=<str>         Filter by tag (#work, #home, etc)

  --today                       Show only tasks due today or earlier
  -u --upcoming                 Show future tasks
  -a --all                      Show all todos

  --sort_by=<key>               Sort by given key or comma-separated keys

  --format=<string>             Format result with given template string.


Todo statuses:
    Todo status is a symbol placed in the square buckets ([x], [ ], [~], etc). You can use
    whatever you want, `dela` doesn't have eny presets, except default `--status=~,n, `,
    which you can overwrite.


Upcoming Tasks:

    If a task has a date in future, it's hidden by default. To see such tasks in the output,
    add the `--upcoming` flag.


Sorting:

    By default todos are sorted by `--status` value according to the order of given statuses,
    and then by date to place todos with dates to the top of the list (due date sounds important,
    right?). You can sort todos by passing an attribute name to the  `--sort_by` flag.
    Possible values are:

    - title
    - date
    - status
    - tags
    - file


Formatting output:

    You can use your own template with the `--format` flag. Unicode colors are supported:

    For example:

        dela list --format='- [$status] $file: $title'

    Next variables are supported:

    - $title
    - $date
    - $status
    - $tags
    - $file
    - $line

Frontmatter Support:

    You can use frontmatter to override file-level todo states, it can be useful in the next cases:
        - Automatically consider all todos in `todos_done.md` as done
        - To set `date` as `today` automatically if a todo is placed in `today.md`
        - To automatically set per-file tags

    To achieve this, you can add a frontmatter in a manner like below:

    ```
    ---
    dela:
        date: today # it's a specific keyword. Standard dates in a YYYYmmdd format also supported
        status: 'x' # override status for all todos in the file
        tags: # append per-file tags
            - per_file_tag
    ---
    ```

Examples:

    dela list # By default prints all todos with [~], [n] and [ ] checkboxes
    dela list -s=' ,s' # prints todos only with statuses [ ] and [s]
    dela list --today # prints only todos with default statuses with a date equal or earliear then today
    dela list -t=#tag # prints only todos that have tag #tag
    dela list -u -s='~,n, ,s' #prints todos with given status including upcoming ones
"""

version = pkg_resources.get_distribution('dela').version


def main():
    args = docopt(doc, version=version)

    if args['--verbose']:
        log.setLevel('INFO')
        log.info('Verbose logging has been enabled')

    log.info(f'{args}')

    if args['list'] == True:
        config = ListCommandConfig(args)

        result = ListCommand(config).run()
        presentation = TodoPresentation(config.format)
        presentation.present(result)
        exit(0)

if __name__ == '__main__':
    main()



import os, glob
from dataclasses import dataclass
import frontmatter
from dela.logger import log

@dataclass
class Line:
    file: str
    file_meta: dict
    number: int
    content: str


def get_offset_len(original_content, cropped_content):
    original_len = len(original_content.split('\n'))
    cropped_len = len(cropped_content.split('\n'))
    return original_len - cropped_len


class FileReader:
    @staticmethod
    def get_lines(g):
        files = [file for file in glob.glob(os.path.expanduser(g))]
        log.debug(f'Match files: {files}')

        for file_path in files:
            log.debug(f'Parsing file: {file_path}')

            with open(file_path, 'r') as f:
                try:
                    f_content = f.read()
                    fm = frontmatter.loads(f_content)
                    offset_len = get_offset_len(f_content, fm.content)
                    lines = fm.content.split('\n')

                    for [lnum, line] in enumerate(lines):
                        yield Line(
                            file=file_path,
                            file_meta=fm.metadata.get('dela', {}),
                            number=lnum + offset_len,
                            content=line,
                        )
                except Exception:
                    log.info(f'Broken file: {file_path}')



class ListCommandConfig(object):
    def __init__(self, args):
        self.glob = args['<glob>'] if args['<glob>'] else '*.md'
        self.format = (
            args['--format']
            if args['--format']
            else '\u001b[30m- \u001b[0m\u001b[01m[$status]\u001b[0m \u001b[31m$file:$line\u001b[0m $title \u001b[0m\u001b[34m$tags\u001b[0m \u001b[31m$date\u001b[0m'
        )
        self.all = True if args['--all'] else False
        self.statuses = args['--statuses'].replace('=', '').split(',') if args['--statuses'] else []
        self.tags = args['--tags'].replace('=', '').split(',') if args['--tags'] else []
        self.today = True if args['--today'] else False
        self.upcoming = True if args['--upcoming'] else False
        self.sort_by = args['--sort_by'].replace('=', '').split(',') if args['--sort_by'] else None

    def __str__(self):
        return str(self.__class__) + ': ' + str(self.__dict__)




from dela.todo_query import TodoQuery
from dela.logger import log


class ListCommand:
    def __init__(self, config) -> None:
        self.config = config

    def run(self):
        log.info(f'Execute list command with config: {self.config}')

        query = TodoQuery(self.config.glob).start()

        if not self.config.all:
            query.status_in(self.config.statuses)

        if self.config.tags:
            query.tag_in(self.config.tags)

        if self.config.today:
            query.today()

        if not self.config.upcoming and not self.config.all:
            query.exclude_upcoming()

        if not self.config.all:
            query.sort_by_statuses(self.config.statuses)
            query.sort_by(['date'])

        if self.config.sort_by:
            query.sort_by(self.config.sort_by)

        return query.exec()




import re
from datetime import datetime

TODO_RE = '^\s*-\s+\[(?P<status>.{1})\]\s+(?:(?P<date>\d{8})\s+)?(?P<title>.*?)(?:\s+(?P<tagline>#.*))?$'


class Todo(object):
    title = None
    date = None
    file = None
    tags = []
    status = None
    line = None

    def __init__(self, title, status, tags, file, date, line):
        self.title = title
        self.date = date
        self.status = status
        self.tags = tags
        self.file = file
        self.line = line

    @staticmethod
    def from_line(line, file, lnum):
        match = match = re.search(TODO_RE, line)

        if not match:
            return None

        gd = match.groupdict()

        return Todo(
            title=gd['title'],
            status=gd['status'].lower(),
            date=gd['date'] if 'date' in gd and gd['date'] else '',
            tags=gd['tagline'].split(' ')
            if 'tagline' in gd and gd['tagline'] is not None
            else [],
            file=file,
            line=lnum,
        )

    def merge(self, data):
        if data.get('date'):
            self.date = self._parse_date(data.get('date'))

        if data.get('tags'):
            self.tags = [*data.get('tags'), *self.tags]

        if data.get('status'):
            self.status = data.get('status')

    def _parse_date(self, date):
        if date == 'today':
            return datetime.today().strftime('%Y%m%d')

        return date



    def __repr__(self):
        return f'title: {self.title}\ndate: {self.date}\ntags: {",".join(self.tags)}\nstatus: {self.status}\nfile: {self.file}\nline: {self.line}'
